
(defvar subrec-mode-map ()
  "local keymap for subrec mode buffers")

(defconst subrec-prog-name "subrec" 
  "*Name of program to type check subrec definitions.")


(defconst subrec-process-name "subrec" "*The name of the subrec-process")




(if subrec-mode-map
    ()
  (setq subrec-mode-map (make-sparse-keymap))
 ;; (define-key subrec-mode-map "\r"  'subrec-electric-return)
 ;; (define-key subrec-mode-map "\t" 'subrec-indent-line)
  (define-key subrec-mode-map "\C-c\C-s" 'subrec-pop-to-shell)
  (define-key subrec-mode-map "\C-c\C-u" 'subrec-save-buffer-use-file)
  (define-key subrec-mode-map "\C-c\C-a" 'subrec-add-definition)
  (define-key subrec-mode-map "\C-c\C-d" 'subrec-delete-definition)

  (define-key subrec-mode-map "\C-c\C-g" 'subrec-send-region)
  (define-key subrec-mode-map "\C-c\C-f" 'subrec-run-on-file)
  (define-key subrec-mode-map "\C-c\C-l" 'subrec-run-on-files)
  (define-key subrec-mode-map "\C-c\C-e" 'subrec-show-environment)
  (define-key subrec-mode-map "\C-c\C-c" 'subrec-get-definition)
  (define-key subrec-mode-map "\C-c\C-t" 'subrec-get-type)
  (define-key subrec-mode-map "\C-c\C-w" 'subrec-eval-whnf)
  (define-key subrec-mode-map "\C-c\C-h" 'subrec-help)
  (define-key subrec-mode-map "\C-c\C-x" 'subrec-clear-definitions)
  (define-key subrec-mode-map "\C-c\C-q" 'subrec-quit-system)
  (define-key subrec-mode-map "\C-c\C-r" 'subrec-reload-buffer))


(defun subrec-set-syntax-table ()
  "Set the syntax table for subrec-mode."
  (setq subrec-mode-syntax-table (make-syntax-table))
  (set-syntax-table subrec-mode-syntax-table)
  (modify-syntax-entry ?\' "w")
  (modify-syntax-entry ?_  "w")
  (modify-syntax-entry ?:  "_")
  (modify-syntax-entry ?%  "_ 14")
  (modify-syntax-entry ?- "_ 23")
;;  (modify-syntax-entry ?<  "(>")
;;  (modify-syntax-entry ?>  ")<")
  (set-syntax-table subrec-mode-syntax-table))


;; font-lock

(defconst  subrec-font-lock-keywords (purecopy
  (list
   '("^[ \t]*\\(let\\|in\\|use\\)\\>"
     1 font-lock-doc-string-face)
   '("\\<\\(Type\\|Set\\)\\>"
     1 font-lock-type-face)
   '("^\\(\\sw+\\)[ \t]*[(:]"
     1  font-lock-function-name-face t t )
   '("^\\s-*\\(\\sw+\\)\\s-*:" 1 font-lock-variable-name-face)))
  "Additional expressions to highlight in subrec mode.")

(defun subrec-mode ()
  "Subrec mode.
Major mode to edit definitions for Subrec type-checker.

Commands:
\\{subrec-mode-map}"
  (interactive)
  (kill-all-local-variables)
  (setq mode-popup-menu subrec-menu)
  (set-buffer-menubar current-menubar)
  (add-submenu nil subrec-menu)
  (subrec-set-syntax-table)
  (setq major-mode 'subrec-mode)
  (setq mode-name "SubRec Definitions")
  (make-local-variable 'subrec-shell-visible)
  (setq subrec-shell-visible nil) ;; inferior shell no visible
  (use-local-map subrec-mode-map))


(defconst subrec-menu
  '( "SubRec"
     "---"
    ["Init SubRec type-checker"		subrec-pop-to-shell            	t]
    ["Check this file"                  subrec-save-buffer-use-file     t]
    ["Check this definition"            subrec-add-definition           t]
    ["Check region"                     subrec-send-region              t]
    ["Reload this file"                 subrec-reload-buffer            t]
    ["Load file ..."	        	subrec-run-on-file	       	t]
    ["Load file of files..."	        subrec-run-on-files	       	t]
    ["Show constant"     	        subrec-get-definition	       	t]
    ["Type of constant"	                subrec-get-type 	       	t]
    ["Evaluate"	                        subrec-eval-whnf 	       	t]
    ["Show environment"	        	subrec-show-environment	       	t]
    ["Delete last definition"           subrec-delete-definition        t]
    ["Reset environment"                subrec-clear-definitions        t]
    ["Help "                            subrec-help                     t]
    ["Quit "                            subrec-quit-system              t]
    "---"
    ))


;;
;; Inferior shell
;;

(defvar subrec-shell-map nil "The mode map for subrec-shell.")

(defun subrec-shell ()
  "Inferior shell invoking subrec.
It is not possible to have more than one shell running subrec.
Like the shell mode with the additional command:

\\[subrec-run-on-file]\t Runs haskell on the file.
\\{subrec-shell-map}

Variables controlling the mode:

subrec-prog-name (default \"subrec\")
    The string used to invoke the haskell program.



subrec-process-name (default \"subrec\")
    The name of the process running haskell.

subrec-shell-prompt-pattern (default \"^> *\")
    The prompt pattern.

Runs subrec-shell-hook if not nil."
  (interactive)
  (if (not (process-status subrec-process-name))
      (save-excursion			; Process is not running
	(message "Starting subrec...")	; start up a new process
	(require 'shell)
	(set-buffer 
	      (make-comint subrec-process-name subrec-prog-name))
	(erase-buffer) ; Erase the buffer if a previous
	(if subrec-shell-map		; process died in there
	    ()
	  (setq subrec-shell-map (copy-keymap shell-mode-map))
	  (define-key subrec-shell-map "\C-c\C-f" 'haskell-run-on-file))
;;	  (define-key subrec-shell-map "\C-c\C-f" 'subrec-run-on-file))
	(use-local-map subrec-shell-map)
	(make-local-variable 'shell-prompt-pattern)
	(setq shell-prompt-pattern ">")
	(setq major-mode 'subrec-shell)
	(setq mode-name "SUBREC Shell")
	(set-process-filter (get-process subrec-process-name)
			    'subrec-process-filter)
	(message "Starting Subrec...done.")
	(run-hooks 'subrec-shell-hook))))

(defun subrec-process-filter (proc str)
  (let ((cur (current-buffer))
	(pop-up-windows t))
    (pop-to-buffer (concat "*" subrec-process-name "*"))
    (goto-char (point-max))
    (if (string= str "\b\b\b  \b\b\b")
	(backward-delete-char 4)
      (insert str))
 ;; Estudio si hubo problemas
 ;;   (if (string-match "\\(^Problem:.*\\)$" str)
 ;;	(progn
 ;;	  (message (match-string 1 str))
 ;;	  (beep)))
    (set-marker (process-mark proc) (point-max))
    (pop-to-buffer cur)))
  
(defun subrec-pop-to-shell ()
  (interactive)
  (subrec-shell)
  (pop-to-buffer (concat "*" subrec-process-name "*")))

(defun subrec-run-on-file (fil)
  (interactive "FRun subrec on : ")
  (subrec-shell)
  (save-some-buffers)
  (process-send-string subrec-process-name
	       (concat "load " (base-name fil)  "\n"
		     )))

;; estas las puse yo 07/08

(defun subrec-run-on-files (fil)
  (interactive "FRun subrec on : ")
  (subrec-shell)
  (save-some-buffers)
  (process-send-string subrec-process-name
	       (concat "loadf " (base-name fil)  "\n"
		     )))


(defun subrec-get-definition (cname)
  (interactive "sName of constant : ")
  (subrec-shell)
  (process-send-string subrec-process-name
	       (concat "getdef " cname  "\n"
		     )))

(defun subrec-get-type (cname)
  (interactive "sName of constant : ")
  (subrec-shell)
  (process-send-string subrec-process-name
	       (concat "type " cname  "\n"
		     )))

(defun subrec-eval-whnf (exp)
  (interactive "sExpression : ")
  (subrec-shell)
  (process-send-string subrec-process-name
	       (concat "whnf " exp  "\n"
		     )))

;; 

(defun subrec-save-buffer-use-file ()
  "Save the buffer, and send a `use file' to the inferior shell
running Subrec."
  (interactive)
  (let (file)
    (if (setq file (buffer-file-name))	; Is the buffer associated
	(progn				; with file ?
	  (save-buffer)
	  (subrec-shell)
	  (process-send-string subrec-process-name
		       (concat "load " file "\n")))
;;			       (base-name file)
;;			        "\n")))
      (error "Buffer not associated with file."))))

(defvar subrec-home-dir "/inco/09/usr/gustun/")

(defun base-name (file)
   (concat subrec-home-dir (car (last (split-string file "~")))))
;;  (car (last(split-string file "/"))))

;; estas tambien las agregue yo

(defun subrec-show-environment ()
 "shows the current environment"
  (interactive)
  (subrec-shell)
  (process-send-string subrec-process-name "env \n"))

(defun subrec-delete-definition ()
 "shows the current environment"
  (interactive)
  (subrec-shell)
  (process-send-string subrec-process-name "del \n"))

(defun subrec-help ()
 "shows the help of the system"
  (interactive)
  (subrec-shell)
  (process-send-string subrec-process-name "help \n"))

(defun subrec-quit-system ()
 "quits the system"
  (interactive)
  (subrec-shell)
  (process-send-string subrec-process-name "quit \n"))

;;

(defun subrec-clear-definitions ()
  "Borra todas las definiciones en el ambiente 
del proceso Subrec en ejecucion"
  (interactive)
  (subrec-shell)
  (process-send-string subrec-process-name
		       "reset \n"))

(defun subrec-reload-buffer ()
  (interactive)
  (subrec-clear-definitions)
  (subrec-save-buffer-use-file))


" This below is an attempt of defining a function for checking regions (definitions)"
" Is a modification of the function  send-region from haskell-mode.el "

;; The template used for temporary files, created when a region is
;; send to the inferior process running subrec.

(defconst subrec-tmp-template "/tmp/subrec.tmp."
  "*Template for the temporary file, created by subrec-simulate-send-region.")

(defvar subrec-check-command "load"
  "The command sent to check a region/buffer.")

 (defun subrec-make-temp-name (pre)
  (concat (make-temp-name pre) ".sr"))

 (defvar subrec-tmp-files-list nil
  "List of all temporary files created by haskell-simulate-send-region.
Each element in the list is a list with the format:

      (\"tmp-filename\"  buffer  start-line)")
   
 (defvar subrec-simulate-send-region-called-p nil
  "Has haskell-simulate-send-region been called previously.")

 (defun subrec-simulate-send-region (point1 point2)
  "Simulate send region. As send-region only can handle what ever the
system sets as the default, we have to make a temporary file.
Updates the list of temporary files (subrec-tmp-files-list)."
  (let ((file (expand-file-name (subrec-make-temp-name subrec-tmp-template))))
    ;; Remove temporary files when we leave emacs
    (if (not subrec-simulate-send-region-called-p)
	(progn
	  (add-hook 'kill-emacs-hook 'subrec-remove-tmp-files)
	  (setq subrec-simulate-send-region-called-p t)))
    (save-excursion
      (goto-char point1)
      (setq subrec-tmp-files-list
	    (cons (list file
			(current-buffer)
			(save-excursion	; Calculate line no.
			  (beginning-of-line)
			  (1+ (count-lines 1 (point)))))
		  subrec-tmp-files-list)))
    (write-region point1 point2 file nil 'dummy)
    (subrec-shell)
    (message "Using temporary file: %s" file)
    (send-string
     subrec-process-name
     ;; string to send: load file;
     (concat (concat subrec-check-command " ")
	     file "\n"))))


(defun subrec-remove-tmp-files ()
  "Remove the temporary files, created by haskell-simulate-send-region, if
they still exist. Only files recorded in haskell-tmp-files-list are removed."
  (message "Removing temporary files created by haskell-mode...")
  (while subrec-tmp-files-list
    (condition-case ()
 	(delete-file (car (car subrec-tmp-files-list)))
      (error ()))
    (setq subrec-tmp-files-list (cdr subrec-tmp-files-list)))
  (message "Removing temporary files created by haskell-mode...done."))

(defun subrec-send-region ()
  "Send region."
  (interactive)
  (let (start end)
    (save-excursion
      (setq end (point))
      (exchange-point-and-mark)
      (setq start (point)))
    (subrec-simulate-send-region start end)))

(defun subrec-send-buffer ()
  "Send the buffer."
  (interactive)
  (subrec-simulate-send-region (point-min) (point-max)))


"For checking one definition"

(defun backward-begin-def ()
"Se posiciona al comienzo de una definicion.
Se asume que una definicion empieza en el comienzo de la linea"
  (interactive)
  (search-backward-regexp "^\\sw+"))


(defun forward-end-def ()
"Se posiciona al final de una definicion"
  (interactive)
  (search-forward-regexp "^\\sw+\\|\\'" () t)
  (if (not (looking-at ";"))
       (search-backward ";"))
      )

(defun subrec-add-definition ()
  "Carga la definicion donde el cursor
   esta posicionado. Si esta en el medio de 2 defs.
   o al final del buffer carga la anterior;
   -NO salva el buffer-"
  (interactive)
  (save-excursion
  (let* 
      (;;comienzo de definicion
       (beg (progn			
	      (backward-begin-def)
	      (point)))
       ;;nombre de la constante
       (name-def (progn (looking-at "\\sw+")
			(match-string 0)))
       ;; fin de la definicion 
       (fin (progn			
	      (forward-word 1)
	      (forward-end-def)
	      (point)))
       ;; texto de la definicion, se eliminan los fines de linea
       (def (replace-in-string (buffer-substring beg fin)
			       "\n" " ")))
    ;; Consulta si desea o no cargar la definicion
    ( if (y-or-n-p-maybe-dialog-box 
	  (concat "Add definition of \"" name-def "\"?"))
	(progn 
	  (message (concat "Add definition`" name-def "'...")) 
	  (subrec-shell)
	  (process-send-string subrec-process-name 
		       (concat "add " def "\n"))
	  (message (concat
		    "Add definifion `" name-def "'...Done" ))))
      )))


