module AbsSyntax where

import MyPrelude

data Exp =   
           EIdent Name
         | ECons Name
         | EPar Int Name
	 | EApp Exp Exp
	 | EAbs Name Exp
         | ESel Exp Name
         | ERecO Exp Fields 

         | EPair Exp Exp
         | EFst Exp
         | ESnd Exp

	 | ESort Name
	 | EPi Name Exp Exp
         | ERecT Exp Fields
         | ESigma Name Exp Exp

         | ELet Bool Defs Exp
-- first version         | EUse Exp Exp Exp
         | EUse Exp Coerc Exp
        deriving(Read,Show)

addLabelRT (ERecT ro flds) lb = ERecT ro (flds++[lb])
addLabelRT rt lb = ERecT rt [lb] 

addLabelRO (ERecO ro flds) lb = ERecO ro (flds++[lb])
addLabelRO ro lb = ERecO ro [lb] 



data Coerc = Coerc Exp | Infer deriving(Read,Show)
eqCoerc :: Coerc -> Coerc -> Bool
eqCoerc (Coerc e1) (Coerc e2) = e1 == e2
eqCoerc Infer Infer = True
eqCoerc _ _ = False

instance Eq Coerc where
  (==) = eqCoerc

eqExp :: Exp -> Exp -> Bool
eqExp (EIdent n) (EIdent n') = n == n'
eqExp (ECons c) (ECons c') = c == c'
eqExp (EPar i _) (EPar i' _) = i == i'
eqExp (EApp f a) (EApp g b) = eqExp f g && eqExp a b
eqExp (EAbs n e) (EAbs n' e') = n == n' && eqExp e e'
eqExp (ESel e n) (ESel e' n') = n == n' && eqExp e e'
eqExp (ERecO ro ofld) (ERecO ro' ofld') = eqExp ro ro' && eqFlds ofld ofld'
eqExp (ESort s) (ESort s') = s == s'
eqExp (EPi n alfa beta) (EPi n' alfa' beta') =
    n == n' && eqExp alfa beta && eqExp alfa' beta'
eqExp (ERecT rt tfld) (ERecT rt' tfld') = eqExp rt rt' && eqFlds tfld tfld'
eqExp (EPair f a) (EPair g b) = eqExp f g && eqExp a b
eqExp (EFst e1) (EFst e2) = eqExp e1 e2
eqExp (ESnd e1) (ESnd e2) = eqExp e1 e2
eqExp (ESigma n alfa beta) (ESigma n' alfa' beta') =
    n == n' && eqExp alfa beta && eqExp alfa' beta'
eqExp (EUse e1 e2 e3) (EUse e1' e2' e3') =
    e1 == e1' && e2 == e2' && e3 == e3'
eqExp _ _ = False -- eqExp for lets???

eqFld (n,e) (n',e') = n == n' && eqExp e e'
eqFlds fs1 fs2 = and (zipWith eqFld fs1 fs2) 

instance Eq Exp where 
   (==) = eqExp
 
mkIdent = EIdent
mkParam = EPar
mkCons = ECons
mkApp = EApp
mkAbs = EAbs
mkSel = ESel
mkRecO = ERecO
mkSort = ESort
mkPi = EPi
mkRecT = ERecT
mkLet = ELet
mkSig = ESigma
mkPair = EPair
mkFst = EFst
mkSnd = ESnd
mkUse = EUse

dummyParam i = mkParam i (show i)

emptyRecO = mkRecO (ECons "emptyRO") []
emptyRecT = mkRecT (ECons "emptyRT") []

isemptyRecO r = r == emptyRecO
isemptyRecT r = r == emptyRecT

nonRecursive = False
recursive = True

-- Type Abbreviations

type Name = String
type Sig = (Int,Exp)
type Fields = [(Name,Exp)]
type Labels = [Name]
type Pars = [Int]
type Defs = [Def]

maxPar :: Pars -> Int
maxPar ls | null ls = 0
          | otherwise = maximum ls


-- Ctxt

type Ctxt = [Sig]
paramsOfCtxt c = map fst c
expsOfCtxt c = map snd c

addCtxt = (:)

-- Definition

data Def = DefCons Name Exp
         | Def Name Exp Exp
	 | DefRec Name Exp Exp
         deriving(Eq,Read,Show)

mkDefCons = DefCons
mkDef = Def
mkDefRec = DefRec

isDefCons (DefCons _ _) = True
isDefCons _  = False

isDExpl (Def _ _ _) = True
--isDExpl (DefRec _ _ _) = True
isDExpl _  = False


nameOfDef (DefCons n _) = n
nameOfDef (Def n _ _) = n
nameOfDef (DefRec n _ _) = n

expOfDef (Def _ e _ ) = e
--expOfDef (DefRec _ e _ ) = e
expOfDef _ = error "expOfDef: Impossible case, This is a bug!!!"

typeOfDef (DefCons _ t) = t
typeOfDef (Def _ _ t) = t
typeOfDef (DefRec _ _ t) = t

data Disc = PCons Exp | ExpCons Exp | RecCons deriving(Eq)

getDef (DefCons n _) = PCons $ mkCons n
getDef (Def _ e _) = ExpCons e
getDef (DefRec _ _  _) = RecCons  -- This is for the wired operators!!!
getDef _ = error "getDef, Impossible case"

-- Environment

data Env = Env [Def] deriving(Eq,Show,Read)

mkEnv = Env
nullEnv = Env []
defsOfEnv (Env dfs) = dfs
addDef = (:)
addDefs :: Defs -> Defs -> Defs
addDefs = (++) 
addDefEnv df (Env dfs) = Env (addDef df dfs)
addDefsEnv dfs (Env dfs1) = Env (addDefs dfs dfs1)
nameOfEnv (Env dfs) = map nameOfDef dfs


-- Auxiliar function (for using after parsing)

idtoCons :: [Name] -> [Name] -> Exp -> Exp
idtoCons glob local e =
   case e of
     EIdent n -> if n `elem` local
                 then e
                 else if n `elem` glob
                      then mkParam (read n) n
                      else mkCons n
     ECons _  -> e
     EPar _ _ -> e
     EApp e1 e2 -> EApp (idtoCons glob local e1) (idtoCons glob local e2)
     EAbs x e1 -> EAbs x (idtoCons glob (x:local) e1)
     ESel e x -> ESel (idtoCons glob local e) x
     ERecO e flds  -> ERecO (idtoCons glob local e)
                            [(lb,idtoCons glob local e1) | (lb,e1) <- flds]
                                 
     ESort n -> e
     EPi x alfa beta -> EPi x (idtoCons glob local alfa) 
                              (idtoCons glob (x:local) beta)
     ERecT e flds -> ERecT (idtoCons glob local e)
                           [(lb,idtoCons glob local e1) | (lb,e1) <- flds]
     ELet b dfs e -> ELet b (filterDefsH glob local dfs) (idtoCons glob local e)
     ESigma x alfa beta -> ESigma x (idtoCons glob local alfa) 
                                    (idtoCons glob (x:local) beta)
     EPair e1 e2 -> EPair (idtoCons glob local e1) (idtoCons glob local e2)
     EFst e1 -> EFst (idtoCons glob local e1)
     ESnd e1 -> ESnd (idtoCons glob local e1)
-- fst version  EUse e1 e2 e3 -> EUse (idtoCons glob local e1) (idtoCons glob local e2)(idtoCons glob local e3)
     EUse e1 e2 e3 -> EUse (idtoCons glob local e1) (idtoConsC e2)(idtoCons glob local e3)
                      where idtoConsC (Coerc t) = Coerc (idtoCons glob local t)
                            idtoConsC Infer = Infer
     _ -> error "idtoCons, not defined for this expression"

-- this function will be needed while we are checking judgements with GIVEN 
-- context and environment. 
-- It should be removed when environments are in the state.

filterDef = filterDefH [] []
filterDefs = filterDefsH [] []

filterDefsH :: [Name] -> [Name] -> [Def] -> [Def]
filterDefsH glob loc = map (filterDefH glob loc)

filterDefH ::  [Name] -> [Name] -> Def -> Def
filterDefH glob loc (DefCons n typ) = mkDefCons n (idtoCons glob loc typ)
filterDefH glob loc (Def n term typ) = 
       mkDef n (idtoCons glob loc term) (idtoCons glob loc typ)
filterDefH glob loc (DefRec n term typ) = 
       mkDefRec n (idtoCons glob loc term) (idtoCons glob loc typ)

{-
filterCtxt :: Ctxt -> Ctxt
filterCtxt = snd . (mapAccuml transacc [])
             where transacc acc (p,t) = ((show p) : acc, (p,idtoCons acc [] t))


-- Similar as above, but not quite. Terms and Types will always have to be 
-- filtered (typable terms must be closed)
 
proccessJ :: [Def] -> Ctxt -> Exp -> Exp -> ([Def],Ctxt,Exp,Exp)
proccessJ dfs ctxt term typ =                   
        let pars = map (\(n,_) -> show n) ctxt   {-params coerced to names-}
            dfs' = filterDefs  dfs                {- temporary -}
	    ctxt' = reverse (filterCtxt ctxt)
	    clterm = idtoCons pars [] term       {- closed terms and kinds -}
	    cltype = idtoCons pars [] typ
        in  (dfs',ctxt',clterm,cltype)
            
-}

desParam :: Exp -> Exp
desParam e =
   case e of
     EIdent n -> e 
     ECons _  -> e
     EPar _ x -> EIdent x
     EApp e1 e2 -> EApp (desParam e1) (desParam e2)
     EAbs x e1 -> EAbs x (desParam e1)
     ESel e x -> ESel (desParam e) x
     ERecO e flds  -> ERecO (desParam e)
                            [(lb,desParam e1) | (lb,e1) <- flds]
                                 
     ESort n -> e
     EPi x alfa beta -> EPi x (desParam alfa) 
                              (desParam beta)
     ERecT e flds -> ERecT (desParam e)
                           [(lb,desParam e1) | (lb,e1) <- flds]
     ELet b dfs e -> ELet b (desParamDefs dfs) (desParam e)
     ESigma x alfa beta -> ESigma x (desParam alfa) 
                                    (desParam beta)
     EPair e1 e2 -> EPair (desParam e1) (desParam e2)
     EFst e1 -> EFst (desParam e1)
     ESnd e1 -> ESnd (desParam e1)
     EUse e1 e2 e3 -> EUse (desParam e1) (desParamC e2)(desParam e3)
                      where desParamC (Coerc t) = Coerc (desParam t)
                            desParamC Infer = Infer
     _ -> error "idtoCons, not defined for this expression"

desParamDef ::  Def -> Def
desParamDef (DefCons n typ) = mkDefCons n (desParam typ)
desParamDef (Def n term typ) = 
       mkDef n (desParam term) (desParam typ)
desParamDef (DefRec n term typ) = 
       mkDefRec n (desParam term) (desParam typ)

desParamDefs = map desParamDef
