module CheckMon where

import MyPrelude

import AbsSyntax
import Subst
import Axioms

import NewPPrint

import Exception

import MonadSTE
import CheckState
import MonadCheck

import ConvMon

-- Checking of G |- a : alfa
-- I think that both le and re are valid terms should be a precondition for 
-- this function

checkExp :: Exp -> Exp -> ChM ()
checkExp le re =

    case le  of

          ESort s -> 
            case re of
              ESort t -> do checkAxTypChM (s,t)
              _ -> errorChM "checkExp, Sort expected" 

          EPi x alfa beta ->
            do 
              checkExp alfa re 
              (c,p) <- getCtxtParamChM
              let par   = mkParam p x
                  beta' = substVar par x beta
		  c'    = addCtxt (p,alfa) c
              putCtxtParamChM (c',(p+1))
              checkExp beta' re

          EAbs x e  ->
              do 
                typ <- whnf re
                case typ of
                 EPi y alfa beta ->
                   do
                    (c,p) <- getCtxtParamChM
		    let par1  = mkParam p x
                        par2  = mkParam p y
			e'    = substVar par1 x e
			beta' = substVar par2 y beta
			c'    = addCtxt (p,alfa) c
                    putCtxtParamChM (c',(p+1))
		    checkExp e' beta' 
                 _ -> errorChM "checkExp, Pi type expected" 

-- Extension with records

          ERecT rt tflds ->
	    if re /= minsort
	    then
             if isemptyRecT le then return ()
              else
	       do
	         checkExp rt re
		 prt <- plainRecChM rt
		 let rect   = ERecT emptyRecT prt
		     rtlabs = map fst prt
		 checkRecT tflds rect rtlabs re 
	    else errorChM "checkExp, record type object of first universe??"
				  

          ERecO ro oflds -> 
            do
              typ <- whnf re 
              case typ of
                ERecT rt tflds -> checkRecO (ro,oflds) (rt,tflds) 
                _ -> errorChM "checkExp, Record type expected"

{-
-- Extension with use expressions

          EUse r coerc e1 ->
              do
	        ro <- case coerc of {Coerc ro -> return ro; Infer -> inferType r} 
	        checkExp ro topsort
	        typ <- whnf ro
		case typ of
                  ERecT rt tflds->
                       if (isemptyRecT typ) 
                       then checkExp e1 re
                       else
                          do
			    checkExp r typ
                            prt <- plainRecChM rt
			    let aflds = prt ++ tflds
                            e2 <- mkLetFromUse r aflds [] e1
			    checkExp e2 re
                  _ -> errorChM "checkExp,EUse, Record type expected"
-}

{- This is a still correct definition but more efficient -}

-- Extension with use expressions

          EUse r coerc e1 ->
              do
	        ro <- case coerc of {Coerc ro -> return ro; Infer -> inferType r} 
	        checkExp ro topsort
	        typ <- whnf ro
		case typ of
                  ERecT rt tflds->
                       if (isemptyRecT typ) 
                       then checkExp e1 re
                       else
                          do
			    checkExp r typ
                            prt <- plainRecChM rt
			    let aflds = prt ++ tflds
                            defs <- mkDefsFromUse r aflds [] -- redefine to version with local accumulator!!
                            env <- getEnvChM
			    putEnvChM (addDefsEnv defs env)
			    checkExp e1 re
			    putEnvChM env
                  _ -> errorChM "checkExp,EUse, Record type expected"


-- Extension with let(rec)

          ELet isrec dfs e  ->
	       if isrec
	       then checkLetrec dfs e re 
	       else checkLet dfs e re

-- Extension with sigma types

          ESigma x alfa beta ->
            do 
              (c,p) <- getCtxtParamChM
              checkExp alfa re 
              let par   = mkParam p x
                  beta' = substVar par x beta
		  c'    = addCtxt (p,alfa) c
              putCtxtParamChM (c',(p+1))
              checkExp beta' re

          EPair e1 e2  ->
              do 
                typ <- whnf re
                case typ of
                 ESigma x alfa beta ->
                   do
                    checkExp e1 alfa
		    let beta' = substVar e1 x beta
		    checkExp e2 beta'
                 _ -> errorChM "checkExp, Sigma type expected"

          _ -> do
                 reinf <- inferType le 
                 typeIncl reinf re                 

checkRecT :: Fields -> Exp -> Labels -> Exp -> ChM ()
checkRecT [] _ _ _ = return ()
checkRecT ((lb,fam):rl) rect lbls typ =
  if lb `notElem` lbls
  then
    do
      nfam <- whnf fam
      case nfam of
        EAbs x t -> 
	  do
	   (c,p) <- getCtxtParamChM
	   let par   = dummyParam p
               c'    = addCtxt (p,rect) c
           putCtxtParamChM (c',(p+1))    
	   checkExp (substVar par x t) typ
	   checkRecT rl (addLabelRT rect (lb,nfam)) (lb:lbls) typ -- application of constructors!
	_ -> errorChM ("checkRecT, Type family expected in field " ++ 
                        printBind (lb,fam))
  else
      errorChM ("checkExp, duplicated label in record type: " ++ lb)

{-
mkLetFromUse :: Exp -> Fields -> Defs -> Exp -> ChM Exp -- Ugly, if works redefine!!
mkLetFromUse ro [] defs exp  = return (mkLet nonRecursive defs exp)
mkLetFromUse ro ((lb,fam):rl) defs exp =
    do
      nfam <- whnf fam
      case nfam of
          EAbs x t -> 
              do
	        let df = mkDef lb (mkSel ro lb) (substVar ro x t)
                mkLetFromUse ro rl (df:defs) exp -- The order does no matter, more efficient
          _ -> errorChM "mkLetFromUse, Not a family???!!!"
-}

mkDefsFromUse :: Exp -> Fields -> Defs -> ChM Defs -- Ugly, if works redefine!!
mkDefsFromUse ro [] defs  = return defs 
mkDefsFromUse ro ((lb,fam):rl) defs =
    do
      nfam <- whnf fam
      case nfam of
          EAbs x t -> 
              do
	        let df = mkDef lb (mkSel ro lb) (substVar ro x t)
                mkDefsFromUse ro rl (df:defs) -- The order does no matter, more efficient
          _ -> errorChM "mkLetFromUse, Not a family???!!!"
 
checkRecO :: (Exp,Fields) -> (Exp,Fields) -> ChM ()
checkRecO (ro,oflds) (rt,tflds) =
  do
   let r = mkRecO ro oflds
       noflds = length oflds
       ntflds = length tflds
   if noflds == ntflds
    then if isemptyRecO ro && isemptyRecT rt 
           then checkPlRec oflds tflds r
	   else do 
                  checkExp ro rt 
                  checkPlRec oflds tflds r
    else
        if noflds < ntflds
         then
           if isemptyRecO ro
            then errorChM "checkRecO, record type larger than record object"
            else
	     do
	       let (stflds',stflds'') = splitAt (ntflds-noflds) tflds
               checkExp ro (mkRecT rt stflds')
	       checkPlRec oflds stflds'' r
         else
           if isemptyRecT rt
            then errorChM "checkRecO, record object larger than record type"
            else
	     do
	       let (soflds',soflds'') = splitAt (noflds-ntflds) oflds
               checkExp (mkRecO ro soflds') rt
	       checkPlRec soflds'' tflds r

checkPlRec :: Fields -> Fields -> Exp -> ChM ()
checkPlRec [] [] _ = return ()
checkPlRec []  _ _ = errorChM "r.t. larger than r.o."
checkPlRec _  [] _ = errorChM "r.o. larger than r.t."
checkPlRec ((lbo,e):ofs) rect@((lbt,t):tfs) ro =
  if lbo == lbt 
  then 
      do
       nt <- whnf t
       case nt of
         EAbs x t1 -> 
	   do
	     checkExp e (substVar ro x t1)
	     checkPlRec ofs tfs ro 
	 _ -> errorChM ("checkPlRec: bad formed r.t. " ++ printPlainRecT rect)
  else errorChM ("checkPlRec: distinct labels: " 
                  ++ show lbo ++ " /= " ++ show lbt)


inferType :: Exp -> ChM Exp
inferType e =
         case e of 
            EPar n _ -> lookupCtxtChM n
--            ECons c -> lookupEnvChM c
            EIdent n -> lookupEnvChM n
            EApp e1 e2 -> 
               do
                 typ  <- inferType e1 
                 typn <- whnf typ
                 case typn of
                     EPi x alfa beta ->
                       do
                         checkExp e2 alfa 
                         return  (substVar e2 x beta) {- e2 must be closed!! -}
                     _ -> inferError "Pi" e

            ESel ro lb ->
              if dummyLabE ro 
              then
                 errorChM ("inferType, not a valid reference to the label " 
                            ++ lb)
              else 
	       do
	         typ <- inferType ro
		 typn <- whnf typ
		 case typn of
		   ERecT rt tflds ->
		      do
		        prt <- plainRecChM rt
			typlb <- lookupFldsChM (prt ++ tflds) lb
			case typlb of
			   EAbs x beta -> return (substVar ro x beta)
			   _ -> inferError "type family" e

                   _ -> inferError "Record" e

            EFst e1 ->
               do
		 typ  <- inferType e1 
                 typn <- whnf typ
                 case typn of
                     ESigma x alfa beta -> return alfa
                     _ -> inferError "Sigma" e

            ESnd e1 ->
               do
		 typ  <- inferType e1 
                 typn <- whnf typ
                 case typn of
                     ESigma x alfa beta -> return (substVar (EFst e1) x beta)
                     _ -> inferError "Sigma" e

       
            _ -> errorChM ("inferType, cannot infer type of expression: " 
                           ++ show e)

dummyLabE e = e == (EIdent "dummylab") || e == (ECons "dummylab")
             
inferError str e = 
  errorChM mssg
  where mssg = "inferType," ++ str ++ " type expected " ++
               "when inferring type of  " ++ printExp e

-- Let(rec) checking                           

checkLet :: Defs -> Exp -> Exp -> ChM ()
checkLet dfs e typ =
   do
    chst <- getChM
    checkDefs dfs
    checkExp e typ
    putChM chst
    
checkLetrec dfs e typ = 
   do
    chst <- getChM
    putEnvChM (addDefsEnv dfs (envOfChSt chst))
    checkDefs dfs
    checkExp e typ
    putChM chst

defTrace :: Def -> ChM ()
defTrace d =
  do
    let mssg = printCheckDef d 
    myTrace mssg (checkDef d)


{- This function has the side-effect of extending the environment -}

checkDef :: Def -> ChM ()
checkDef d@(DefCons n e) =
   do
     env <- getEnvChM
     checkExp e topsort
     putEnvChM (addDefEnv d env)
     
checkDef d@(Def n exp typ) =
   do
     env <- getEnvChM
     checkExp typ topsort
     checkExp exp typ
     putEnvChM (addDefEnv d env)

checkDef d@(DefRec n exp typ) =
   do
     env <- getEnvChM
     checkExp typ topsort
     putEnvChM (addDefEnv (DefRec n exp typ) env) -- mmmhhh!!!!
     checkExp exp typ
     putEnvChM (addDefEnv d env)

checkDef _ = errorChM "checkDef, impossible case"

{- This inherits the side-effect of checkDef -}

checkDefs :: Defs -> ChM ()
checkDefs [] = return ()
checkDefs (d:ds) =
  do
    defTrace d 
    checkDefs ds
