module CheckState where

import AbsSyntax
import Axioms
import Exception
import Env

type CheckState = (Ctxt,Env,Int,AxiomTyp,AxiomIncl)

ctxtOfChSt (c,_,_,_,_) = c
envOfChSt  (_,e,_,_,_) = e
paramOfChSt (_,_,p,_,_) = p
axTypOfChSt (_,_,_,at,_) = at
axInclOfChSt (_,_,_,_,ai) = ai

getCtxtParamChSt (c,_,p,_,_) = (c,p)

putCtxtChSt c (_,e,p,at,ai) = (c,e,p,at,ai)
putEnvChSt e (c,_,p,at,ai) = (c,e,p,at,ai)
putParamChSt p (c,e,_,at,ai) = (c,e,p,at,ai)
putAxTypChSt at (c,e,p,_,ai) = (c,e,p,at,ai)
putAxInclChSt ai (c,e,p,at,_) = (c,e,p,at,ai)

putCtxtParamChSt (c,p) (_,e,_,at,ai) = (c,e,p,at,ai)

lookupChC :: CheckState -> Int -> E Exp
lookupChC = lookupC . ctxtOfChSt 

lookupChE :: CheckState -> Name -> E Exp
lookupChE = lookupEnv . envOfChSt

getAllDefChSt :: CheckState -> Name -> E Def
getAllDefChSt = getAllDefEnv . envOfChSt

delLastDefChSt :: CheckState -> CheckState
delLastDefChSt (c,e,p,at,ai) = (c,delLastDefEnv e,p,at,ai)

incrPar :: CheckState -> CheckState
incrPar (c,e,p,at,ai) = (c,e,p+1,at,ai)

isAxTyp ::  (Name,Name) -> CheckState -> E ()
isAxTyp e chst = 
      if e `elem` (axTypOfChSt chst)
      then return ()
      else errorE ("isAxTyp, not a typing axiom" ++ show e)

isAxInc ::  (Name,Name) -> CheckState -> E ()
isAxInc e chst = 
      if e `elem` (axInclOfChSt chst)
      then return ()
      else errorE ("isAxIncl, not a type inclusion  axiom" ++ show e)

initState ctxt env = (ctxt,env,0,initAxTyp,initAxIncl)
