module ConvMon where

import MyPrelude

import AbsSyntax
import Subst
import Axioms

import NewPPrint

import Exception
import Env 

import MonadSTE 
import CheckState 
import MonadCheck 


typeIncl :: Exp -> Exp -> ChM ()
typeIncl typ1 typ2 =
  do
    typ1n <- whnf typ1 
    typ2n <- whnf typ2 
    if typ1n == typ2n 
       then return ()
       else    
           case (typ1n,typ2n) of
              
           (ESort s1,ESort s2) -> checkAxInclChM (s1,s2)

	   (EPi x alfa1 beta1, EPi y alfa2 beta2) ->
               do
		 (c,p) <- getCtxtParamChM
                 typeIncl alfa2 alfa1
		 let par1 = mkParam p x
                     par2 = mkParam p y
		     beta1' = substVar par1 x beta1
		     beta2' = substVar par2 y beta2
		     c' = addCtxt (p,alfa2) c
		 putCtxtParamChM    (c',(p+1))
		 typeIncl  beta1' beta2'

           (ERecT rho1 sig1, ERecT rho2 sig2) ->
	       do
	         prho1 <- plainRecChM rho1
	         prho2 <- plainRecChM rho2
		 let apr1 = prho1++sig1
		     apr2 = prho2++sig2
		 (c,p) <- getCtxtParamChM
		 let par = dummyParam p
		     c' = addCtxt (p,ERecT emptyRecT apr1) c
		 putCtxtParamChM (c',p+1)
		 inclSig apr1 apr2 par

	   (ESigma x alfa1 beta1, ESigma y alfa2 beta2) ->
               do
		 (c,p) <- getCtxtParamChM
                 typeIncl alfa1 alfa2
		 let par1 = mkParam p x
                     par2 = mkParam p y
		     beta1' = substVar par1 x beta1
		     beta2' = substVar par2 y beta2
		     c' = addCtxt (p,alfa1) c
		 putCtxtParamChM    (c',(p+1))
		 typeIncl  beta1' beta2'

           (EApp _ _, EApp _ _) -> termConv typ1 typ2 minsort {- hmmmmm!!! -}

           _ -> errorChM 
                     ("typeIncl, types are not in the inclusion relation " 
                      ++ "\n" ++ printExp typ1n ++ " and\n " ++ printExp typ2n)


inclSig _ [] _ = return ()
inclSig s1 ((lb,f2):flds) p =
  do
    f1 <- updateErrorChM (lookupFldsChM s1 lb)"in typeIncl"
    typeIncl (EApp f1 p) (EApp f2 p)
    inclSig s1 flds p
{-
convTrace :: Exp -> Exp -> Exp -> ChM ()
convTrace e1 e2 typ =
   do
     let mssg = printEqual (e1,e2,typ)
     myTrace mssg (termConv e1 e2 typ)
-}

termConv :: Exp -> Exp -> Exp -> ChM ()
termConv e1 e2 t =
     if e1 == e2 
     then return ()
     else
          do
	    typ <- whnf t
            case typ of
               EPi x alfa beta ->
                   do
                     (ctxt,p) <- getCtxtParamChM
		     let par = mkParam p x
		         e1' = EApp e1 par             
		         e2' = EApp e2 par             
		         beta' = substVar par x beta
		         ctxt' = addCtxt (p,alfa) ctxt
		     putCtxtParamChM (ctxt',(p+1))
		     termConv e1' e2' beta'

               ERecT ret sig ->
	          do
		    pret <- plainRecChM ret
		    robjConv e1 e2 (pret++sig)

               ESigma x alfa beta ->
                   do
		     let fst1 = mkFst e1
		         fst2 = mkFst e2
			 snd1 = mkSnd e1
			 snd2 = mkSnd e2
			 beta' = substVar fst1 x beta
		     termConv fst1 fst2 alfa
		     termConv snd1 snd2 beta'

               _ -> 
                   do
                     ne1 <- whnf e1
		     ne2 <- whnf e2 
		     typ1 <- headConv ne1 ne2 
                     typeIncl typ1 typ


robjConv _ _ [] = return ()
robjConv ro1 ro2 ((lb,f):flds) =
  do
    termConv (ESel ro1 lb) (ESel ro2 lb) (EApp f ro1)
    robjConv ro1 ro2 flds
{-
headConvTrace :: Exp -> Exp -> ChM Exp
headConvTrace e1 e2 =
   do
     let mssg = printExp e1 ++ " = " ++ printExp e2
     myTrace mssg (headConv e1 e2)
-}		    

headConv ::  Exp -> Exp -> ChM Exp
headConv e1 e2 =
        case (e1,e2) of
            (EPar m i, EPar n i') ->
                 if m == n 
                 then lookupCtxtChM  m 
                 else errorChM ("headConv, EPar, " ++ i ++ "=?" ++ i')

            (ECons m, ECons n) ->
                 if m == n 
                 then lookupEnvChM  m 
                 else errorChM "headConv, ECons, not head convertibles"

            (EApp f a, EApp g b) ->
                do
                  typ <- headConv f g 
		  ntyp <- whnf typ
                  case ntyp of
                       EPi x alfa beta ->
                         do
                           termConv a b alfa 
                           return (substVar a x beta)    {- a must be closed -}
                       _ -> errorChM "headConv,EApp, Pi expected"

            (ESel ro1 lb1, ESel ro2 lb2) ->
	       if lb1 == lb2 
	       then
	        do
	          typ <- headConv ro1 ro2
		  ntyp <- whnf typ
		  case ntyp of
		    ERecT rt tflds ->
		      do
		       prt <- plainRecChM rt
		       f <- lookupFldsChM (prt++tflds) lb1
		       return (EApp f ro1)
	       else
	          errorChM ("headConv,ESel, distinct labels " 
                             ++ lb1 ++ " =? " ++ lb2)

            (EFst p1, EFst p2) ->
               do
                typ <- headConv p1 p2
		ntyp <- whnf typ
                case ntyp of
                  ESigma x alfa beta -> return alfa
                  _ -> errorChM ("headConv,EFst, Sigma expected "
                                ++ printExp e1 ++ " =? " ++ printExp e2)

            (ESnd p1, ESnd p2) ->
               do
                typ <- headConv p1 p2
		ntyp <- whnf typ
                case ntyp of
                  ESigma x alfa beta -> return (substVar (mkFst p1) x beta)
                  _ -> errorChM ("headConv,ESnd, Sigma expected "
                                ++ printExp e1 ++ " =? " ++ printExp e2)

            _ -> errorChM ("headConv, "++ printExp e1 ++ " =? " ++ printExp e2)



whnf :: Exp -> ChM Exp
whnf e1 = do { env <- getEnvChM; e2 <- reduc e1; putEnvChM env; return e2}

{-
reducTrace :: Exp -> ChM Exp
reducTrace e1 =
   do
     let mssg = "normalising " ++ printExp e1 
     myTrace mssg (reduc e1)
		    
-}

reduc :: Exp -> ChM Exp
reduc e =
  case e of
     EIdent n -> deltaRed n  -- errorChM ("reduc, Ident case" ++ n)
     EPar i n -> return (EPar i n)
     ECons c -> return (ECons c)
     EAbs x e1 -> return (EAbs x e1)
     EApp f a -> betaRed f a
     ESort s -> return (ESort s)
     EPi x alfa beta -> return (EPi x alfa beta)
     ERecT ret sig -> return (ERecT ret sig)
     ERecO reo str -> return (ERecO reo str)
     ESel reo lab -> selred reo lab
     ELet b dfs e1 -> 
           do 
             e2 <- letred dfs e1 
	     reduc e2
     ESigma x alfa beta -> return (ESigma x alfa beta)
     EPair e1 e2 -> return (EPair e1 e2)
     EFst e1 -> fstred e1
     ESnd e1 -> sndred e1
     EUse e1 e2 e3 -> usered e1 e2 e3
     _ -> errorChM "reduc, impossible case" 

letred :: Defs -> Exp -> ChM Exp
letred [] e = return e
letred (d:ds) e = 
  if (isDefCons d) 
  then letred ds e
  else
     do
       let x = nameOfDef d
       de <- reduc (expOfDef d)
       (ds',e') <- mapSubst x de ds e
       letred ds' e'

{- This should be local, but I have problems with the layout of where in do!! -}

mapSubst x de [] e = return ([],substVar de x e)
mapSubst x de (d1:ds1) e =
 do
   let d1' = substVarDef de x d1
       x1 = nameOfDef d1
   if x == x1
      then return (d1':ds1,e)
      else
      do
        (ds2,e1) <- mapSubst x de ds1 e
	return (d1':ds2,e1)

betaRed :: Exp -> Exp -> ChM Exp
betaRed f a =
  do
    f' <- reduc f 
    case f' of
        EAbs x e -> reduc (substVar a x e)       {- a must be closed!! -}
        _ -> return (EApp f' a)

deltaRed :: Name -> ChM Exp
deltaRed n = 
    do
      d <- getConsDefChM n
      case d of
          PCons e -> return e
          ExpCons e -> reduc e
          _ -> errorChM "deltaRed, impossible case"
    
selred :: Exp -> Name -> ChM Exp
selred ro lb =
   if isemptyRecO ro 
   then errorChM "selred, impossible case"
   else
     do
      nro <- reduc ro
      case nro of
        ERecO ero str ->
         do {e <- lookupFldsChM str lb ; reduc e} 
         `escapeChM` selred ero lb 

               {- chst <- getChM
                  elimE (stripChM (lookupFldsChM str lb) chst)
                     (reduc)
                     (\_ -> selred ero lb) -}

	_ -> return (ESel nro lb)

fstred :: Exp -> ChM Exp
fstred e =
   do
     ne <- reduc e
     case ne of
      EPair e1 e2 -> reduc e1
      _ -> return (EFst ne)


sndred :: Exp -> ChM Exp
sndred e =
   do
     ne <- reduc e
     case ne of
      EPair e1 e2 -> reduc e2
      _ -> return (ESnd ne)

{- First version 
usered :: Exp -> Exp -> Exp -> ChM Exp
usered r ro e =
  do
    plr <- plainRecChM ro
    let e1 = foldr (\(lb,_) -> substVar (mkSel r lb) lb) e plr
    reduc e1 -}
   
usered :: Exp -> Coerc -> Exp -> ChM Exp
usered r c e =
    case c of
      Coerc ro ->
         do 
           plr <- plainRecChM ro
	   let e1 = foldr (\(lb,_) -> substVar (mkSel r lb) lb) e plr
	   reduc e1
      Infer -> errorChM "usered, Not yet implemented use with inference"
      _ -> errorChM "usered, Impossible case!!!"

-- Auxiliar function

plainRecChM :: Exp -> ChM Fields
plainRecChM e =
  do
   ne <- reduc e
   if isemptyRecO ne || isemptyRecT ne 
    then return []
    else
        case ne of
	  ERecO e1 oflds -> 
	    do
	      e1fs <- plainRecChM e1
	      return (e1fs ++ oflds)
	  ERecT e1 tflds ->
	    do
	      e1fs <- plainRecChM e1
	      return (e1fs ++ tflds)
          _ -> errorChM "plainRecChM, not a record type"
