module MonadCheck where

import AbsSyntax
import Env
import MonadSTE
import CheckState

type ChM a = STE CheckState a

errorChM = errorSTE
runChM = runSTE
stripChM = stripSTE
updateErrorChM = updateErrorSTE
getChM = getSTE
putChM = putSTE
escapeChM = escapeSTE

-- accessor functions

selectChM :: (CheckState -> a) -> ChM a
selectChM acc =
  do
    chst <- getChM
    return (acc chst)

getCtxtChM = selectChM ctxtOfChSt

getEnvChM = selectChM envOfChSt

getParamChM = selectChM paramOfChSt

getAxTypChM = selectChM axTypOfChSt

getAxInclChM = selectChM axInclOfChSt


-- updating functions

updateChM :: (a -> CheckState -> CheckState) -> a -> ChM ()
updateChM f = updateSTE . f 
 

putCtxtChM = updateChM putCtxtChSt

putEnvChM = updateChM putEnvChSt

putParamChM = updateChM putParamChSt

putAxTypChM = updateChM putAxTypChSt

putAxInclChM = updateChM putAxInclChSt

-- auxiliar functions

getCtxtParamChM = 
  do
   (c,p) <- selectChM getCtxtParamChSt
   putParamChM (p+1)                   -- gensym
   return (c,p)

putCtxtParamChM = updateChM putCtxtParamChSt


-- lookup functions

--lookupChM :: (CheckState -> a -> E Exp) -> a -> ChM Exp
lookupChM f a =
  do
    chst <- getSTE
    liftEtoSTE (f chst a)

lookupCtxtChM :: Int -> ChM Exp
lookupCtxtChM = lookupChM lookupChC

lookupEnvChM :: Name -> ChM Exp
lookupEnvChM = lookupChM lookupChE

-- checking of typing and inclusion axioms

checkAxChM getax str p = 
  do
    al <-  getax
    if p `elem` al
        then return ()
        else errorChM (str ++ ",  pair is not in axioms")


checkAxTypChM :: (Name,Name) -> ChM ()
checkAxTypChM = checkAxChM getAxTypChM "checkAxTypChM"

checkAxInclChM :: (Name,Name) -> ChM ()
checkAxInclChM =  checkAxChM getAxInclChM "checkAxInclChM"


-- unfolding of constant definition

getConsDefChM :: Name -> ChM Disc
getConsDefChM n = 
  do
    env <- getEnvChM
    liftEtoSTE (getConsDefE env n) 

lookupFldsChM :: Fields -> Name -> ChM Exp
lookupFldsChM [] lb = errorChM ("lookupFldsChM: label not found" ++ show lb)
lookupFldsChM ((l,e):flds) lb =
   if l == lb 
   then return e
   else lookupFldsChM flds lb 
