module MonadSTE where

import Exception

data STE s a = STE (s -> E (a,s))


mkSTE = STE
funOfSTE (STE f) = f

instance Monad (STE s) where
   (STE f) >>= g  =  STE (\s -> f s >>= \(a,s') -> funOfSTE (g a) s')
   return         =  \a -> mkSTE (\s -> return (a,s)) 


-- primitive functions  (not abstract)

errorSTE :: String -> STE s a
errorSTE str = mkSTE (\s -> errorE str)

runSTE :: STE s a -> s -> E (a,s)
runSTE (STE f) s = f s
 
stripSTE :: STE s a -> s -> E a
stripSTE (STE f) s = f s >>= (return . fst)

getSTE :: STE s s
getSTE = mkSTE (\s -> return (s,s))

putSTE :: s -> STE s ()
putSTE s = mkSTE (\_ -> return ((),s))

liftEtoSTE :: E a -> STE s a
liftEtoSTE e = mkSTE (\s -> e >>= (\a -> return (a,s)))


liftEfunSTE :: (s -> E a) -> STE s a
liftEfunSTE f = mkSTE (\s -> f s >>= (\a -> return (a,s)))

liftFunSTE :: (s -> a) -> STE s a
liftFunSTE f = mkSTE (\s -> return (f s,s))


updateErrorSTE :: STE s a -> String -> STE s a
updateErrorSTE (STE f) str = 
        mkSTE (\s -> updateErrorE (f s) str) 

escapeSTE :: STE s a -> STE s a -> STE s a
escapeSTE m k = 
   mkSTE (\s -> let STE f = m 
                    STE g = k
                in escapeE (f s) (g s))

-- general functions  (abstract)

appendSTE :: STE s [a]  -> STE s [a]  -> STE s [a] 
appendSTE m k = m >>= (\as ->
                k >>= (\as1 -> return (as++as1)))

concatSTE :: [STE s [a] ] -> STE s [a] 
concatSTE = foldr appendSTE (return [])

zipSTE :: [a] -> [b] -> STE s [(a,b)]
zipSTE as bs = liftEtoSTE (zipE as bs)

updateSTE :: (s -> s) -> STE s () 
updateSTE f = getSTE >>= (putSTE . f)

skipSTE :: STE s ()
skipSTE = updateSTE id

zipWithSTE :: (a -> b -> STE s c) -> [a] -> [b] -> STE s [c] 
zipWithSTE z (a:as) (b:bs) = z a b `consSTE` zipWithSTE z as bs
zipWithSTE _     _      _  = return []

consSTE :: STE s a -> STE s [a] -> STE s [a]
consSTE m k = m >>= (\a ->
              k >>= (\as -> return (a:as)))

sequenceSTE :: [STE s a] -> STE s [a]
sequenceSTE [] = return []
sequenceSTE (k:ks) = k `consSTE` (sequenceSTE ks) 


