module NewPPrint(printExp, printDef, printCheck,printEqual,
                 printBind,printPlainRecT,printEnv,printCheckDef) 

where

import Pretty
import PrettyPrint
import AbsSyntax

pparen False x = x
pparen True  x = parens x 

arrow = text "->"
dot = char '.'
equal = char '='
angles p = char '<' <> p <> char '>'

ppiSeqSepAbove :: String -> (a -> Doc) -> [a] -> Doc
ppiSeqSepAbove _ _ [] = empty
ppiSeqSepAbove sep pf xs = foldr1 (\a b -> (a <> text sep) $$$ b) $ map pf xs

--******* Instances for expressions, definitions,etc *******--

{- The class PPrint is used for taking care of parentheses nesting -}

class PPrint a where
 pPrint :: Int -> a -> Doc

instance PPrint Exp where

   pPrint p (EIdent i) = text i
   pPrint p (ECons n) = text n
   pPrint p (EPar i n) = text n
   pPrint p (ESort s) = text s
   pPrint p (EApp e1 e2) = 
        pparen (p>9) (pPrint 0 e1 <+> pPrint 10 e2 )
   pPrint p (EAbs n e) = 
        pparen (p>0) (text "\\" <> text n <+> arrow <+> pPrint 0 e)
   pPrint p (ELet b ds e) = 
               pparen (p>0)  (kwd <+> braces sdefs) $$ 
                             (text "in  " <> pPrint 0 e)
               where kwd = text (if b then "letrec" else "let")
                     sdefs = ppiSeqSepAbove ";" (pPrint 0) ds 

{- first version    pPrint p (EUse e1 e2 e3) = 
               pparen (p>0)  (kwd <+> (pPrint 0 e1) <+> colon <+> (pPrint 0 e2)) $$ 
                             (text "in  " <> pPrint 0 e3)
               where kwd = text "use"
-}
   pPrint p (EUse e1 e2 e3) = 
               pparen (p>0)  (kwd <+> (pPrint 0 e1) <+> (pPrint 0 e2)) $$ 
                             (text "in  " <> pPrint 0 e3)
               where kwd = text "use"

   pPrint p (EPi n t1  t2) = 
               pparen (p > 0) 
                      (pparen True (text n <+> colon <+> pPrint 0 t1) 
                                    <+> arrow <+> pPrint 0 t2)
   pPrint p (ESel e n) = pparen (p > 0) (pPrint 0 e <> dot <> text n)
   pPrint p (ERecO ro oflds) =
            if isemptyRecO ro 
            then pparen (p > 0) (angles toflds)
            else
                pparen (p > 0) (angles ((pPrint 0 ro <> comma) $$ toflds))
            where toflds = ppiSeqSepAbove "," (pPrint 0 . PrOFld) oflds

   pPrint p (ERecT rt tflds) =
            if isemptyRecT rt
            then pparen (p > 0) (angles ttflds)
            else
                 pparen (p > 0) (angles ((pPrint 0 rt <> comma) $$ ttflds))
            where ttflds = ppiSeqSepAbove "," (pPrint 0 . PrTFld) tflds

   pPrint p (ESigma n t1 t2) =
              pparen (p > 0) (braces (text n <+> colon <+> pPrint 0 t1)
                              <+> pPrint 0 t2)
   pPrint p (EPair e1 e2) =
              pparen (p > 0) 
                     (pparen True (pPrint 0 e1 <> comma <> pPrint 0 e2))

   pPrint p (EFst e1) = pparen (p > 0) (text "fst" <+> pPrint 0 e1)
                     
   pPrint p (ESnd e1) = pparen (p > 0) (text "snd" <+> pPrint 0 e1)
                     
instance PPrint Coerc where
   pPrint p (Coerc e) = colon <+> (pPrint p e)
   pPrint p Infer = text ""
 
instance PPrint Def where
   pPrint p (DefCons n e) = text n <+> colon <+> pPrint p e
   pPrint p (Def n e typ) = 
              (text n <+> colon <+> pPrint p typ) <+>
              (equal <+> pPrint p e)
   pPrint p (DefRec n e typ) = 
              (text "rec" <+> text n <+> colon <+> pPrint p typ)
-- not for wired               <+> equal <+> pPrint p e

instance PPrint Env where
   pPrint _ (Env dfs) = ppiSeqSepAbove ";" (pPrint 0) (reverse dfs)

data PrintOFld = PrOFld (Name,Exp) deriving (Show)
data PrintTFld = PrTFld (Name,Exp) deriving (Show)
data PrintEqual = PrEqu (Exp,Exp,Exp) deriving (Show)
data PrintCheck = PrChk (Exp,Exp) deriving (Show)
data PrintCheckDef = PrChkDef Def deriving (Show)

instance PPrint PrintOFld where
  pPrint p (PrOFld (n,e)) = text n <+> equal <+> pPrint 0 e

instance PPrint PrintTFld where
  pPrint p (PrTFld (n,e)) = text n <+> colon <+> pPrint 0 e

instance PPrint PrintEqual where
  pPrint p (PrEqu (e1,e2,typ)) = 
      pPrint 0 e1 <+> equal <+> pPrint 0 e2 <+> colon <+> pPrint 0 typ

instance PPrint PrintCheck where
  pPrint p (PrChk (e1,e2)) = cat [text "checking ",
                                  pPrint 0 e1, colon <+> pPrint 0 e2]

instance PPrint PrintCheckDef where
  pPrint p (PrChkDef d) = cat [text "checking DEFINITION ",
                               pPrint 0 d]

{- Instances for class Printable -}

instance Printable Exp where
  ppi = pPrint 0

instance Printable Def where
  ppi = pPrint 0

instance Printable Env where
  ppi = pPrint 0

instance Printable PrintOFld where
  ppi = pPrint 0

instance Printable PrintTFld where
  ppi = pPrint 0

instance Printable PrintEqual where
  ppi = pPrint 0

instance Printable PrintCheck where
  ppi = pPrint 0

instance Printable PrintCheckDef where
  ppi = pPrint 0

-- Pretty printing functions, these are the exported ones

printExp :: Exp -> String
printExp = pp

printDef :: Def -> String
printDef = pp

printEqual :: (Exp,Exp,Exp) -> String
printEqual = pp . PrEqu

printCheck :: (Exp,Exp) -> String
printCheck = pp . PrChk

printBind :: (Name,Exp) -> String
printBind = pp . PrTFld

printPlainRecT :: Fields -> String
printPlainRecT = pp . (ERecT emptyRecT)

printEnv :: Env -> String
printEnv = pp

printCheckDef :: Def -> String
printCheckDef = pp . PrChkDef
