{-
  Author: boquist
  Problem: Lot's of datatypes need to be prettyprinted.
  We would like to just 'pp x' to prettyprint 'x'
  and don't have to care about the type of x.
  Solution:
-}
module PrettyPrint(module Pretty,
                   Printable(..),
		   ppiSeq, ppiSeqF, ppiSeqAbove)

where

import Pretty

class Show a => Printable a where
    pp      :: a   -> String               -- Prettyprint
    ppi     :: a   -> Doc                  -- Prettyprint Intelligently
    ppiList :: [a] -> Doc                  -- Specialized for Char
    wrap    :: a   -> Doc		   -- for bracketing


    pp = renderStyle Style { lineLength = 80,
			     ribbonsPerLine = 2.0,
			     mode = PageMode }
	 . ppi

    ppi = text . show

    ppiList [] = text "[]"
    ppiList xs = text "[ " <> ppiSeq ", " xs <> text " ]"

    wrap x = text "(" <> ppi x <> text ")"


-- Predefined instances:

instance Printable Char where
    ppi	c   = text [c]
    ppiList = text
    wrap    = ppi

instance Printable Bool where
    wrap = ppi

instance Printable Int where
    wrap = ppi

instance Printable Integer where
    wrap = ppi

instance (Printable a) => Printable [a] where
    ppi  = ppiList
    wrap = ppi

instance (Printable a, Printable b) => Printable (a, b) where
    ppi (ann, expr) = text "(" <> ppi ann <> text ", " <> ppi expr <> text ")"
    wrap = ppi

instance (Printable a, Printable b, Printable c) => Printable (a, b, c) where
    ppi (a, b, c) = text "(" <> ppi a <> text ", "
                             <> ppi b <> text ", "
			     <> ppi c <> text ")"
    wrap = ppi


ppiSeq :: (Printable a) => String -> [a] -> Doc
ppiSeq _   [] = empty
ppiSeq sep xs = foldr1 (\a b -> a <> text sep <> b) $ map ppi xs


ppiSeqF :: String -> (a -> Doc) -> [a] -> Doc
ppiSeqF _   _  [] = empty
ppiSeqF sep pf xs = foldr1 (\a b -> a <> text sep <> b) $ map pf xs


ppiSeqAbove :: (a -> Doc) -> [a] -> Doc
ppiSeqAbove _  [] = empty
ppiSeqAbove pf xs = foldr1 ($$) $ map pf xs


