import MyPrelude

import AbsSyntax
import Axioms
import Exception

import NewPPrint

import MonadSTE
import CheckState
import MonadCheck

import ConvMon
import CheckMon

import Lex
import ParserMonad
import Parser

import ParseCom
import System
import IO

interProb p t = do 
                  putStrLn ("Problem: "++ t)
		  loop p

analyse :: String -> State -> IO ()
analyse c st =
 if null c then loop st
 else
  elimE (readLoad (lexanal c))
        (\cm -> case cm of
	             Load s -> loadFile st s

                     LoadF s -> loadFiles st s 

                     PrintEnv -> printIO st (printEnv (envOfSt st))

                     Add d -> contEIO (runChM (checkDef d) st)
			              (\(_,st1) -> loop st1)
				      (interProb st)

                     WhNf e ->  contEIO (runChM (whnf e) st)
                                        (\(e1,_) -> do putStrLn (printExp e1)
                                                       loop st)
                                        (interProb st)

                     DelLast -> delLastDef st

                     GetDef dn -> contEIO (getAllDefSt st dn)
                                          (\def -> do putStrLn (printDef def)
					              loop st)
					  (interProb st)

                     TypeOf n -> contEIO (lookupChE st n)
                                         (\typ -> do putStrLn(printBind(n,typ))
					             loop st)
                                         (interProb st)

                     ParseE e -> printIO st (printExp e)

                     ParseD d -> printIO st (printDef d)
				    
		     Reset -> loop initialState

                     Help  -> printHelp st
                     
                     Quit -> return ()
        )

        (interProb st)



type State = CheckState  -- (only checking complete terms)
putEnvSt = putEnvChSt
getAllDefSt = getAllDefChSt
initialState = initState [] nullEnv :: State
envOfSt = envOfChSt
delLastDefSt = delLastDefChSt

loop :: State -> IO ()
loop state 
        = do putStr ">> "
             inp <- getLine
             analyse inp state                                                

main :: IO ()
main = do {putStrLn welcmssg; loop initialState}

-- Commands

delLastDef = loop . delLastDefSt

contEIO :: E a -> (a -> IO b) -> (String -> IO b) -> IO b
contEIO = elimE  

-- Printing 

printIO :: State -> String -> IO ()
printIO st s = do {putStrLn s ; loop st}


-- I/O Commands 

welcmssg = "Version 0.0.4 (records, let and sigma), " 
           ++ "type help for an online help"

safeReadFile :: String -> IO String
safeReadFile fn = 
    catch (readFile fn) (\_ -> return "not found")

readFileAndDo :: String -> State -> (String -> IO ()) -> IO () 
readFileAndDo fn st success =
  do
   ans <- safeReadFile fn
   if filenotfound ans
    then interProb st (fn ++ " " ++ ans)
    else success ans
     
filenotfound = (== "not found")


checkfileIO :: State -> (State -> IO ()) -> String -> IO ()
checkfileIO st stf fcts =
  elimE (checkFileE st fcts)
        (\(_,st1) -> stf st1)
	(interProb st)   

checkFileE st s =  
  let STE f = do ds <- liftEtoSTE (parseDefs s)
                 checkDefs ds
		 putEnvChM (addDefsEnv (reverse ds) (envOfSt st))
  in f st

loadFile :: State -> String -> IO ()
loadFile st fn =
   readFileAndDo fn st (checkfileIO st loop)

loadFiles :: State -> String -> IO ()
loadFiles st fn = 
   readFileAndDo fn st (\fcts -> readFiles st (lines fcts))
   where readFiles st [] = loop st
         readFiles st (l:ls) = 
                 readFileAndDo l st (checkfileIO st (\st1 -> readFiles st1 ls))

printHelp :: State -> IO ()
printHelp st = 
  do
   let prhelp fcts = do {mapM_ putStrLn (lines fcts);loop st}   
   homedir <- getEnv "HOME"
   readFileAndDo (homedir ++ "/SUBREC/help") st prhelp


