module Subst where

import AbsSyntax

-- Substitution of an Exp for a variable in an Exp ([x:=e]f)
-- does not prevent capture, but respects hiding

substVar :: Exp -> Name -> Exp -> Exp
substVar e x f =
  case f of
    EIdent y -> if y == x then e else f
    ECons c -> ECons c
    EPar i n -> EPar i n
    EApp e1 e2 -> EApp (substVar e x e1) (substVar e x e2)
    EAbs y f1 -> if y == x 
                 then EAbs y f1 
                 else EAbs y (substVar e x f1)
    ESel e1 lb -> ESel (substVar e x e1) lb
    ERecO ro flds -> ERecO (substVar e x ro) 
                           [(lb,substVar e x v) | (lb,v) <- flds] 
    ESort _ -> f
    EPi y alfa beta -> if y == x 
                       then EPi y alfa beta 
                       else EPi y (substVar e x alfa) (substVar e x beta)
    ERecT rt flds  -> ERecT (substVar e x rt) 
                            [(lb,substVar e x v) | (lb,v) <- flds] 
--    ELet b dfs e1 -> ELet b (substVarDefs e x dfs) (substVar e x e1)
    ELet b defs e1 -> 
            substVarDefs defs []
            where
                substVarDefs [] l  = mkLet b (reverse l) (substVar e x e1)
                substVarDefs (df:dfs) l =
                     let df' = substVarDef e x df 
		     in if (nameOfDef df) == x
                        then mkLet b ((reverse l) ++ [df'] ++ dfs) e1
                        else substVarDefs dfs (df':l)  
    ESigma y alfa beta -> if y == x 
                          then ESigma y alfa beta 
                          else ESigma y (substVar e x alfa) (substVar e x beta)
    EPair e1 e2 -> EPair (substVar e x e1) (substVar e x e2)
    EFst e1 -> EFst (substVar e x e1)
    ESnd e1 -> ESnd (substVar e x e1)
-- fst version    EUse e1 e2 e3 -> EUse (substVar e x e1) (substVar e x e2) (substVar e x e3)
    EUse e1 e2 e3 -> EUse (substVar e x e1) (substVarC e2) (substVar e x e3)
                     where substVarC (Coerc t) = Coerc (substVar e x t)
		           substVarC Infer = Infer
    _ -> error "substVar, Not defined for this expression"

substVarDef e x (DefCons n typ) = DefCons n (substVar e x typ)
substVarDef e x (Def n exp typ) = 
          Def n (substVar e x exp) (substVar e x typ)
substVarDef e x (DefRec n exp typ) = 
          DefRec n (substVar e x exp) (substVar e x typ) 

--substVarDefs e x = map (substVarDef e x)

substLab :: Name -> Exp -> Exp
substLab n e = case e of
                  EIdent id -> if isdummyLab id
		               then mkIdent n
			       else e
		  EApp e1 e2 -> EApp (substLab n e1) (substLab n e2)
		  EAbs n1 e1 -> EAbs n1 (substLab n e1)
                  EPi x e1 e2 -> EPi x (substLab n e1) (substLab n e2)
		  ESel e1 n1 -> ESel (substLab n e1) n1
		  ERecT ert sig -> ERecT ert (substLabs n sig)
		  ERecO ero str -> ERecO ero (substLabs n str)
		  ELet b dfs e1 -> ELet b (substLabDefs n dfs) (substLab n e1)
                  ESigma x e1 e2 -> ESigma x (substLab n e1) (substLab n e2)
		  EPair e1 e2 -> EPair (substLab n e1) (substLab n e2)
		  EFst e1 -> EFst (substLab n e1)
		  ESnd e1 -> ESnd (substLab n e1)
-- fst version    EUse e1 e2 e3 -> EUse (substLab n e1)(substLab n e2)(substLab n e3)
                  EUse e1 e2 e3 -> EUse (substLab n e1)(substLabC e2)(substLab n e3)
                                   where substLabC (Coerc t) = Coerc (substLab n t)
				         substLabC Infer = Infer
	          _ -> e

isdummyLab i = i == "dummylab"
substLabs n = map (\(n1,e)-> (n1,substLab n e))

substLabDef l (DefCons n typ) = DefCons n (substLab l typ)
substLabDef l (Def n exp typ) = Def n (substLab l exp) (substLab l typ)
substLabDef l (DefRec n exp typ) = DefRec n (substLab l exp) (substLab l typ)

substLabDefs l = map (substLabDef l)


